"""jobflow-remote is a python package to run jobflow workflows on remote resources"""

from jobflow_remote._version import __version__
from jobflow_remote.config.jobconfig import set_run_config
from jobflow_remote.config.manager import ConfigManager
from jobflow_remote.config.settings import JobflowRemoteSettings
from jobflow_remote.jobs.jobcontroller import JobController
from jobflow_remote.jobs.store import get_jobstore
from jobflow_remote.jobs.submit import submit_flow

SETTINGS = JobflowRemoteSettings()



from __future__ import annotations

import datetime
import glob
import logging
import os
import subprocess
import time
import traceback
from pathlib import Path

from jobflow import JobStore, initialize_logger
from jobflow.core.flow import get_flow
from jobflow.core.job import Job
from monty.os import cd
from monty.serialization import dumpfn, loadfn
from monty.shutil import decompress_file

from jobflow_remote.jobs.batch import LocalBatchManager
from jobflow_remote.jobs.data import IN_FILENAME, OUT_FILENAME
from jobflow_remote.remote.data import (
    default_orjson_serializer,
    get_job_path,
    get_remote_store_filenames,
)
from jobflow_remote.utils.log import initialize_remote_run_log

logger = logging.getLogger(__name__)


def run_remote_job(run_dir: str | Path = "."):
    """Run the job"""

    initialize_remote_run_log()

    start_time = datetime.datetime.utcnow()
    with cd(run_dir):
        error = None
        try:
            dumpfn({"start_time": start_time}, OUT_FILENAME)
            in_data = loadfn(IN_FILENAME)

            job: Job = in_data["job"]
            store = in_data["store"]

            # needs to be set here again since it does not get properly serialized.
            # it is possible to serialize the default function before serializing, but
            # avoided that to avoid that any refactoring of the
            # default_orjson_serializer breaks the deserialization of old Fireworks
            store.docs_store.serialization_default = default_orjson_serializer
            for additional_store in store.additional_stores.values():
                additional_store.serialization_default = default_orjson_serializer

            store.connect()

            initialize_logger()
            try:
                response = job.run(store=store)
            finally:
                # some jobs may have compressed the FW files while being executed,
                # try to decompress them if that is the case.
                decompress_files(store)

            # The output of the response has already been stored in the store.
            response.output = None

            # Convert to Flow the dynamic responses before dumping the output.
            # This is required so that the response does not need to be
            # deserialized and converted by to Flows by the runner.
            if response.addition:
                response.addition = get_flow(response.addition)
            if response.detour:
                response.detour = get_flow(response.detour)
            if response.replace:
                response.replace = get_flow(response.replace)

            output = {
                "response": response,
                "error": error,
                "start_time": start_time,
                "end_time": datetime.datetime.utcnow(),
            }
            dumpfn(output, OUT_FILENAME)
        except Exception:
            # replicate the dump to catch potential errors in
            # serializing/dumping the response.
            error = traceback.format_exc()
            output = {
                "response": None,
                "error": error,
                "start_time": start_time,
                "end_time": datetime.datetime.utcnow(),
            }
            dumpfn(output, OUT_FILENAME)


def run_batch_jobs(
    base_run_dir: str | Path,
    files_dir: str | Path,
    process_uuid: str,
    max_time: int | None = None,
    max_wait: int = 60,
    max_jobs: int | None = None,
):
    initialize_remote_run_log()

    # TODO the ID should be somehow linked to the queue job
    bm = LocalBatchManager(files_dir=files_dir, process_id=process_uuid)

    t0 = time.time()
    wait = 0
    sleep_time = 10
    count = 0
    while True:
        if max_time and max_time < time.time() - t0:
            logger.info("Stopping due to max_time")
            return

        if max_wait and wait > max_wait:
            logger.info(
                f"No jobs available for more than {max_wait} seconds. Stopping."
            )
            return

        if max_jobs and count >= max_jobs:
            logger.info(f"Maximum number of jobs reached ({max_jobs}). Stopping.")
            return

        job_str = bm.get_job()
        if not job_str:
            time.sleep(sleep_time)
            wait += sleep_time
        else:
            wait = 0
            count += 1
            job_id, index = job_str.split("_")
            index = int(index)
            logger.info(f"Starting job with id {job_id} and index {index}")
            job_path = get_job_path(job_id=job_id, index=index, base_path=base_run_dir)
            try:
                with cd(job_path):
                    result = subprocess.run(
                        ["bash", "submit.sh"],
                        check=True,
                        text=True,
                        capture_output=True,
                    )
                    if result.returncode:
                        logger.warning(
                            f"Process for job with id {job_id} and index {index} finished with an error"
                        )
                bm.terminate_job(job_id, index)
            except Exception:
                logger.error(
                    "Error while running job with id {job_id} and index {index}",
                    exc_info=True,
                )
            else:
                logger.info(f"Completed job with id {job_id} and index {index}")


def decompress_files(store: JobStore):
    file_names = [OUT_FILENAME]
    file_names.extend(get_remote_store_filenames(store))

    for fn in file_names:
        # If the file is already present do not decompress it, even if
        # a compressed version is present.
        if os.path.isfile(fn):
            continue
        for f in glob.glob(fn + ".*"):
            decompress_file(f)



from __future__ import annotations

import logging
import subprocess
from enum import Enum
from pathlib import Path
from string import Template
from xmlrpc.client import Fault

import psutil
from monty.os import makedirs_p
from supervisor import childutils
from supervisor.states import RUNNING_STATES, STOPPED_STATES, ProcessStates
from supervisor.xmlrpc import Faults

from jobflow_remote.config import ConfigManager, Project

logger = logging.getLogger(__name__)


supervisord_conf_str = """
[unix_http_server]
file=$sock_file

[rpcinterface:supervisor]
supervisor.rpcinterface_factory = supervisor.rpcinterface:make_main_rpcinterface

[supervisord]
logfile=$log_file
logfile_maxbytes=10MB
logfile_backups=5
loglevel=$loglevel
pidfile=$pid_file
nodaemon=$nodaemon

[supervisorctl]
serverurl=unix://$sock_file

[program:runner_daemon]
priority=100
command=jf -p $project runner run -pid -log $loglevel
autostart=true
autorestart=false
numprocs=1
process_name=run_jobflow%(process_num)s
stopwaitsecs=86400
"""


supervisord_conf_str_split = """
[unix_http_server]
file=$sock_file

[rpcinterface:supervisor]
supervisor.rpcinterface_factory = supervisor.rpcinterface:make_main_rpcinterface

[supervisord]
logfile=$log_file
logfile_maxbytes=10MB
logfile_backups=5
loglevel=$loglevel
pidfile=$pid_file
nodaemon=$nodaemon

[supervisorctl]
serverurl=unix://$sock_file

[program:runner_daemon_checkout]
priority=100
command=jf -p $project runner run -pid --checkout -log $loglevel
autostart=true
autorestart=false
numprocs=1
process_name=run_jobflow_checkout
stopwaitsecs=86400

[program:runner_daemon_transfer]
priority=100
command=jf -p $project runner run -pid --transfer -log $loglevel
autostart=true
autorestart=false
numprocs=$num_procs_transfer
process_name=run_jobflow_transfer%(process_num)s
stopwaitsecs=86400

[program:runner_daemon_queue]
priority=100
command=jf -p $project runner run -pid --queue -log $loglevel
autostart=true
autorestart=false
numprocs=1
process_name=run_jobflow_queue
stopwaitsecs=86400

[program:runner_daemon_complete]
priority=100
command=jf -p $project runner run -pid --complete -log $loglevel
autostart=true
autorestart=false
numprocs=$num_procs_complete
process_name=run_jobflow_complete%(process_num)s
stopwaitsecs=86400
"""


class DaemonError(Exception):
    pass


class DaemonStatus(Enum):
    SHUT_DOWN = "SHUT_DOWN"
    STOPPED = "STOPPED"
    STOPPING = "STOPPING"
    PARTIALLY_RUNNING = "PARTIALLY_RUNNING"
    RUNNING = "RUNNING"


class DaemonManager:
    conf_template_single = Template(supervisord_conf_str)
    conf_template_split = Template(supervisord_conf_str_split)

    def __init__(
        self,
        daemon_dir: str | Path,
        log_dir: str | Path,
        project: Project,
    ):
        self.project = project
        self.daemon_dir = Path(daemon_dir).absolute()
        self.log_dir = Path(log_dir).absolute()

    @classmethod
    def from_project(cls, project: Project):
        daemon_dir = project.daemon_dir
        log_dir = project.log_dir
        return cls(daemon_dir, log_dir, project)

    @classmethod
    def from_project_name(cls, project_name: str | None = None):
        config_manager = ConfigManager()
        project = config_manager.get_project(project_name)
        return cls.from_project(project)

    @property
    def conf_filepath(self) -> Path:
        return self.daemon_dir / "supervisord.conf"

    @property
    def pid_filepath(self) -> Path:
        return self.daemon_dir / "supervisord.pid"

    @property
    def log_filepath(self) -> Path:
        return self.log_dir / "supervisord.log"

    @property
    def sock_filepath(self) -> Path:
        path = self.daemon_dir / "s.sock"
        if len(str(path)) > 97:
            msg = f"socket path {path} is too long for UNIX systems. Set the daemon_dir value in the project configuration so that the socket path is shorter"
            raise DaemonError(msg)
        return path

    def clean_files(self):
        self.pid_filepath.unlink(missing_ok=True)
        self.sock_filepath.unlink(missing_ok=True)

    def get_interface(self):
        env = {
            "SUPERVISOR_SERVER_URL": f"unix://{str(self.sock_filepath)}",
            "SUPERVISOR_USERNAME": "",
            "SUPERVISOR_PASSWORD": "",
        }
        interface = childutils.getRPCInterface(env)
        return interface

    def get_supervisord_pid(self) -> int | None:
        pid_fp = self.pid_filepath

        if not pid_fp.is_file():
            return None

        try:
            with open(pid_fp) as f:
                pid = int(f.read().strip())
        except ValueError:
            logger.warning(f"The pid file {pid_fp} could not be parsed")
            return None
        return pid

    def check_supervisord_process(self) -> bool:
        pid = self.get_supervisord_pid()

        running = True
        if pid is None:
            running = False

        try:
            process = psutil.Process(pid)

            for cmdline_element in process.cmdline():
                if cmdline_element.endswith("supervisord"):
                    break
            else:
                running = False

            if process.username() != psutil.Process().username():
                logger.warning(
                    f"pid {pid} is running supervisord, but belongs to a different user"
                )
                running = False
        except psutil.NoSuchProcess:
            running = False

        if not running:
            if pid is not None:
                logger.warning(
                    f"Process with pid {pid} is not running but daemon files are present. Cleaning them up."
                )
            self.clean_files()

        return running

    def check_status(self) -> DaemonStatus:
        process_active = self.check_supervisord_process()

        if not process_active:
            return DaemonStatus.SHUT_DOWN

        if not self.sock_filepath.is_socket():
            raise DaemonError(
                "the supervisord process is alive, but the socket is missing"
            )

        interface = self.get_interface()
        try:
            proc_info = interface.supervisor.getAllProcessInfo()
        except Fault as e:
            # catch this exception as it may be raised if the status is queried while
            # the supervisord process is shutting down. The error is quite cryptic, so
            # replace with one that is clearer. Also see a related issue in supervisord:
            # https://github.com/Supervisor/supervisor/issues/48
            if e.faultString == "SHUTDOWN_STATE":
                raise DaemonError(
                    "The daemon is likely shutting down and the actual state cannot be determined"
                )
            raise
        if not proc_info:
            raise DaemonError(
                "supervisord process is running but no daemon process is present"
            )

        if all(pi.get("state") in RUNNING_STATES for pi in proc_info):
            return DaemonStatus.RUNNING

        if any(pi.get("state") in RUNNING_STATES for pi in proc_info):
            return DaemonStatus.PARTIALLY_RUNNING

        if all(pi.get("state") in STOPPED_STATES for pi in proc_info):
            return DaemonStatus.STOPPED

        if all(
            pi.get("state") in (ProcessStates.STOPPED, ProcessStates.STOPPING)
            for pi in proc_info
        ):
            return DaemonStatus.STOPPING

        raise DaemonError("Could not determine the current status of the daemon")

    def get_processes_info(self) -> dict[str, dict] | None:
        process_active = self.check_supervisord_process()

        if not process_active:
            return None

        pids = {"supervisord": {"pid": self.get_supervisord_pid(), "state": "RUNNING"}}

        if not self.sock_filepath.is_socket():
            raise DaemonError(
                "the supervisord process is alive, but the socket is missing"
            )

        interface = self.get_interface()
        proc_info = interface.supervisor.getAllProcessInfo()
        if not proc_info:
            raise DaemonError(
                "supervisord process is running but no daemon process is present"
            )

        for pi in proc_info:
            pids[pi.get("name")] = {"pid": pi.get("pid"), "state": pi.get("statename")}

        return pids

    def write_config(
        self,
        num_procs_transfer: int = 1,
        num_procs_complete: int = 1,
        single: bool = True,
        log_level: str = "info",
        nodaemon: bool = False,
    ):
        if single:
            conf = self.conf_template_single.substitute(
                sock_file=str(self.sock_filepath),
                pid_file=str(self.pid_filepath),
                log_file=str(self.log_filepath),
                nodaemon="true" if nodaemon else "false",
                project=self.project.name,
                loglevel=log_level,
            )
        else:
            conf = self.conf_template_split.substitute(
                sock_file=str(self.sock_filepath),
                pid_file=str(self.pid_filepath),
                log_file=str(self.log_filepath),
                num_procs_transfer=num_procs_transfer,
                num_procs_complete=num_procs_complete,
                nodaemon="true" if nodaemon else "false",
                project=self.project.name,
                loglevel=log_level,
            )
        with open(self.conf_filepath, "w") as f:
            f.write(conf)

    def start_supervisord(
        self,
        num_procs_transfer: int = 1,
        num_procs_complete: int = 1,
        single: bool = True,
        log_level: str = "info",
        nodaemon: bool = False,
    ) -> str | None:
        makedirs_p(self.daemon_dir)
        makedirs_p(self.log_dir)
        self.write_config(
            num_procs_transfer=num_procs_transfer,
            num_procs_complete=num_procs_complete,
            single=single,
            log_level=log_level,
            nodaemon=nodaemon,
        )
        cp = subprocess.run(
            f"supervisord -c {str(self.conf_filepath)}",
            shell=True,
            capture_output=True,
            text=True,
        )
        if cp.returncode != 0:
            return f"Error staring the supervisord process. stdout: {cp.stdout}. stderr: {cp.stderr}"

        # TODO check if actually started?

        return None

    def start_processes(self) -> str | None:
        interface = self.get_interface()
        result = interface.supervisor.startAllProcesses()
        if not result:
            return "No process started"

        failed = [r for r in result if r.get("status") == Faults.SUCCESS]
        if len(failed) == 0:
            return None
        elif len(failed) != len(result):
            msg = "Not all the daemon processes started correctly. Details: \n"
            for f in failed:
                msg += f"  - {f.get('description')}\n"
            return msg

        return None

    def start(
        self,
        num_procs_transfer: int = 1,
        num_procs_complete: int = 1,
        single: bool = True,
        log_level: str = "info",
        raise_on_error: bool = False,
    ) -> bool:
        status = self.check_status()
        if status == DaemonStatus.RUNNING:
            error = "Daemon process is already running"
        elif status == DaemonStatus.SHUT_DOWN:
            error = self.start_supervisord(
                num_procs_transfer=num_procs_transfer,
                num_procs_complete=num_procs_complete,
                single=single,
                log_level=log_level,
            )
        elif status == DaemonStatus.STOPPED:
            self.shut_down(raise_on_error=raise_on_error)
            error = self.start_supervisord(
                num_procs_transfer=num_procs_transfer,
                num_procs_complete=num_procs_complete,
                single=single,
                log_level=log_level,
            )
            # else:
            #     error = self.start_processes()
        elif status == DaemonStatus.STOPPING:
            error = "Daemon process are stopping. Cannot start."
        else:
            error = f"Daemon status {status} could not be handled"

        if error is not None:
            if raise_on_error:
                raise DaemonError(error)
            else:
                logger.error(error)
                return False
        return True

    def stop(self, wait: bool = False, raise_on_error: bool = False) -> bool:
        status = self.check_status()
        if status in (
            DaemonStatus.STOPPED,
            DaemonStatus.STOPPING,
            DaemonStatus.SHUT_DOWN,
        ):
            return True

        if status == DaemonStatus.RUNNING:
            interface = self.get_interface()
            if wait:
                result = interface.supervisor.stopAllProcesses()
            else:
                result = interface.supervisor.signalAllProcesses(15)

            error = self._verify_call_result(result, "stop", raise_on_error)

            return error is None

        raise DaemonError(f"Daemon status {status} could not be handled")

    def _verify_call_result(
        self, result, action: str, raise_on_error: bool = False
    ) -> str | None:
        error = None
        if not result:
            error = f"The action {action} was not applied to the processes"
        else:
            failed = [r for r in result if r.get("status") == Faults.SUCCESS]
            if len(failed) != len(result):
                error = f"The action {action} was not applied to all the processes. Details: \n"
                for f in failed:
                    error += f"  - {f.get('description')}\n"

        if error is not None:
            if raise_on_error:
                raise DaemonError(error)
            else:
                logger.error(error)
                return error

        return None

    def kill(self, raise_on_error: bool = False) -> bool:
        status = self.check_status()
        if status == DaemonStatus.SHUT_DOWN:
            logger.info("supervisord is not running. No process is running")
            return True

        if status in (DaemonStatus.RUNNING, DaemonStatus.STOPPING):
            interface = self.get_interface()
            result = interface.supervisor.signalAllProcesses(9)
            error = self._verify_call_result(result, "kill", raise_on_error)

            return error is None

        raise DaemonError(f"Daemon status {status} could not be handled")

    def shut_down(self, raise_on_error: bool = False) -> bool:
        status = self.check_status()
        if status == DaemonStatus.SHUT_DOWN:
            logger.info("supervisord is already shut down.")
            return True
        interface = self.get_interface()
        try:
            interface.supervisor.shutdown()
        except Exception:
            if raise_on_error:
                raise
            return False
        return True



def test_version():
    from jobflow_remote import __version__

    assert __version__.startswith("0.0.1")


def test_imports():
    """This test triggers all the top-level imports by importing
    the global `SETTINGS`.

    """
    from jobflow_remote import SETTINGS  # noqa

    ...



from __future__ import annotations

import io
import logging
import shlex
import traceback
from pathlib import Path

import fabric
from paramiko.ssh_exception import SSHException

from jobflow_remote.remote.host.base import BaseHost

logger = logging.getLogger(__name__)


class RemoteHost(BaseHost):
    """
    Execute commands on a remote host.
    For some commands assumes the remote can run unix
    """

    def __init__(
        self,
        host,
        user=None,
        port=None,
        config=None,
        gateway=None,
        forward_agent=None,
        connect_timeout=None,
        connect_kwargs=None,
        inline_ssh_env=None,
        timeout_execute=None,
        keepalive=60,
        shell_cmd="bash",
        login_shell=True,
        retry_on_closed_connection=True,
    ):
        self.host = host
        self.user = user
        self.port = port
        self.config = config
        self.gateway = gateway
        self.forward_agent = forward_agent
        self.connect_timeout = connect_timeout
        self.connect_kwargs = connect_kwargs
        self.inline_ssh_env = inline_ssh_env
        self.timeout_execute = timeout_execute
        self.keepalive = keepalive
        self.shell_cmd = shell_cmd
        self.login_shell = login_shell
        self.retry_on_closed_connection = retry_on_closed_connection
        self._create_connection()

    def _create_connection(self):
        self._connection = fabric.Connection(
            host=self.host,
            user=self.user,
            port=self.port,
            config=self.config,
            gateway=self.gateway,
            forward_agent=self.forward_agent,
            connect_timeout=self.connect_timeout,
            connect_kwargs=self.connect_kwargs,
            inline_ssh_env=self.inline_ssh_env,
        )

    def __eq__(self, other):
        if not isinstance(other, RemoteHost):
            return False
        return self.as_dict() == other.as_dict()

    @property
    def connection(self):
        return self._connection

    def execute(
        self,
        command: str | list[str],
        workdir: str | Path | None = None,
        timeout: int | None = None,
    ):
        """Execute the given command on the host

        Parameters
        ----------
        command: str or list of str
            Command to execute, as a str or list of str.
        workdir: str or None
            path where the command will be executed.

        Returns
        -------
        stdout : str
            Standard output of the command
        stderr : str
            Standard error of the command
        exit_code : int
            Exit code of the command.
        """
        self._check_connected()

        if isinstance(command, (list, tuple)):
            command = " ".join(command)

        # TODO: check if this works:
        if not workdir:
            workdir = "."

        workdir = Path(workdir)

        timeout = timeout or self.timeout_execute

        if self.shell_cmd:
            shell_cmd = self.shell_cmd
            if self.login_shell:
                shell_cmd += " -l "
            shell_cmd += " -c "
            remote_command = shell_cmd + shlex.quote(command)
        else:
            remote_command = command

        with self.connection.cd(workdir):
            out = self._execute_remote_func(
                self.connection.run,
                remote_command,
                hide=True,
                warn=True,
                timeout=timeout,
            )

        return out.stdout, out.stderr, out.exited

    def mkdir(
        self, directory: str | Path, recursive: bool = True, exist_ok: bool = True
    ) -> bool:
        """Create directory on the host."""
        directory = Path(directory)
        command = f"mkdir {'-p ' if recursive else ''}{str(directory)!r}"
        try:
            stdout, stderr, returncode = self.execute(command)
            if returncode != 0:
                logger.warning(
                    f"Error creating folder {directory}. stdout: {stdout}, stderr: {stderr}"
                )
            return returncode == 0
        except Exception:
            logger.warning(f"Error creating folder {directory}", exc_info=True)
            return False

    def write_text_file(self, filepath: str | Path, content: str):
        """Write content to a file on the host."""
        self._check_connected()

        f = io.StringIO(content)

        self._execute_remote_func(self.connection.put, f, str(filepath))

    def connect(self):
        self.connection.open()
        if self.keepalive:
            self.connection.transport.set_keepalive(self.keepalive)

    def close(self) -> bool:
        try:
            self.connection.close()
        except Exception:
            return False
        return True

    @property
    def is_connected(self) -> bool:
        return self.connection.is_connected

    def put(self, src, dst):
        self._check_connected()

        self._execute_remote_func(self.connection.put, src, dst)

    def get(self, src, dst):
        self._check_connected()

        self._execute_remote_func(self.connection.get, src, dst)

    def copy(self, src, dst):
        cmd = ["cp", str(src), str(dst)]
        self.execute(cmd)

    def _execute_remote_func(self, remote_cmd, *args, **kwargs):
        if self.retry_on_closed_connection:
            try:
                return remote_cmd(*args, **kwargs)
            except OSError as e:
                msg = getattr(e, "message", str(e))
                error = e
                if "Socket is closed" not in msg:
                    raise e
            except SSHException as e:
                error = e
                msg = getattr(e, "message", str(e))
                if "Server connection dropped" not in msg:
                    raise e
            except EOFError as e:
                error = e
        else:
            return remote_cmd(*args, **kwargs)

        # if the code gets here one of the errors that could be due to drop of the
        # connection occurred. Try to close and reopen the connection and retry
        # one more time
        logger.warning(
            f"Error while trying to execute a command on host {self.host}:\n"
            f"{''.join(traceback.format_exception(error))}"
            "Probably due to the connection dropping. "
            "Will reopen the connection and retry."
        )
        try:
            self.connection.close()
        except Exception:
            logger.warning(
                "Error while closing the connection during a retry. "
                "Proceeding with the retry.",
                exc_info=True,
            )
        self._create_connection()
        self.connect()
        return remote_cmd(*args, **kwargs)

    def listdir(self, path: str | Path):
        self._check_connected()

        try:
            return self._execute_remote_func(self.connection.sftp().listdir, str(path))
        except FileNotFoundError:
            return []

    def remove(self, path: str | Path):
        self._check_connected()

        self._execute_remote_func(self.connection.sftp().remove, str(path))

    def _check_connected(self) -> bool:
        """
        Helper method to determine if fabric consider the connection open and
        open it otherwise

        Since many operations requiring connections happen in the runner,
        if the connection drops there are cases where the host may not be
        reconnected. To avoid this issue always try to reconnect automatically
        if the connection is not open.

        Returns
        -------
        True if the connection is open.
        """

        if not self.is_connected:
            # Note: raising here instead of reconnecting demonstrated to be a
            # problem for how the queue managers are handled in the Runner.
            self.connect()
        return True



from __future__ import annotations

import io
import logging
import os
from pathlib import Path
from typing import Any

import orjson
from jobflow.core.store import JobStore
from maggma.stores.mongolike import JSONStore
from monty.json import jsanitize

from jobflow_remote.utils.data import uuid_to_path


def get_job_path(
    job_id: str, index: int | None, base_path: str | Path | None = None
) -> str:
    if base_path:
        base_path = Path(base_path)
    else:
        base_path = Path()

    relative_path = uuid_to_path(job_id, index)
    return str(base_path / relative_path)


def get_remote_in_file(job, remote_store):
    d = jsanitize(
        {"job": job, "store": remote_store},
        strict=True,
        allow_bson=True,
        enum_values=True,
    )
    return io.BytesIO(orjson.dumps(d, default=default_orjson_serializer))


def default_orjson_serializer(obj: Any) -> Any:
    type_obj = type(obj)
    if type_obj != float and issubclass(type_obj, float):
        return float(obj)
    raise TypeError


def get_remote_store(
    store: JobStore, launch_dir: str | Path, add_orjson_serializer: bool = True
) -> JobStore:
    serialization_default = None
    if add_orjson_serializer:
        serialization_default = default_orjson_serializer

    docs_store = JSONStore(
        os.path.join(launch_dir, "remote_job_data.json"),
        read_only=False,
        serialization_default=serialization_default,
    )
    additional_stores = {}
    for k in store.additional_stores.keys():
        additional_stores[k] = JSONStore(
            os.path.join(launch_dir, f"additional_store_{k}.json"),
            read_only=False,
            serialization_default=serialization_default,
        )
    remote_store = JobStore(
        docs_store=docs_store,
        additional_stores=additional_stores,
        save=store.save,
        load=store.load,
    )

    return remote_store


def get_remote_store_filenames(store: JobStore) -> list[str]:
    filenames = ["remote_job_data.json"]
    for k in store.additional_stores.keys():
        filenames.append(f"additional_store_{k}.json")

    return filenames


def update_store(store: JobStore, remote_store: JobStore, db_id: int):
    try:
        store.connect()
        remote_store.connect()

        additional_stores = set(store.additional_stores.keys())
        additional_remote_stores = set(remote_store.additional_stores.keys())

        # This checks that the additional stores in the two stores match correctly.
        # It should not happen if not because of a bug, so the check could maybe be
        # removed
        if additional_stores ^ additional_remote_stores:
            raise ValueError(
                f"The additional stores in the local and remote JobStore do not "
                f"match: {additional_stores ^ additional_remote_stores}"
            )

        # copy the data store by store, not using directly the JobStore.
        # This avoids the need to deserialize the store content and the "save"
        # argument.
        for add_store_name, remote_add_store in remote_store.additional_stores.items():
            add_store = store.additional_stores[add_store_name]

            for d in remote_add_store.query():
                data = dict(d)
                data.pop("_id", None)
                add_store.update(data)
        main_docs_list = list(remote_store.docs_store.query({}))
        if len(main_docs_list) > 1:
            raise RuntimeError(
                "The downloaded output store contains more than one document"
            )
        main_doc = main_docs_list[0]
        main_doc.pop("_id", None)
        # Set the db_id here and not directly in the Job's metadata to prevent
        # it from being propagated to its children/replacements.
        if "db_id" not in main_doc["metadata"]:
            main_doc["metadata"]["db_id"] = db_id
        store.docs_store.update(main_doc, key=["uuid", "index"])
    finally:
        try:
            store.close()
        except Exception:
            logging.error(f"error while closing the store {store}", exc_info=True)
        try:
            remote_store.close()
        except Exception:
            logging.error(
                f"error while closing the remote store {remote_store}", exc_info=True
            )


def resolve_job_dict_args(job_dict: dict, store: JobStore) -> dict:
    """
    Resolve the references in a serialized Job.

    Similar to Job.resolve_args, but without the need to deserialize the Job.
    The references are resolved inplace.

    Parameters
    ----------
    job_dict
        The serialized version of a Job.
    store
        The JobStore from where the references should be resolved.

    Returns
    -------
        The updated version of the input dictionary with references resolved.
    """
    from jobflow.core.reference import OnMissing, find_and_resolve_references

    on_missing = OnMissing(job_dict["config"]["on_missing_references"])
    cache: dict[str, Any] = {}
    resolved_args = find_and_resolve_references(
        job_dict["function_args"], store, cache=cache, on_missing=on_missing
    )
    resolved_kwargs = find_and_resolve_references(
        job_dict["function_kwargs"], store, cache=cache, on_missing=on_missing
    )
    resolved_args = tuple(resolved_args)

    # substitution is in place
    job_dict["function_args"] = resolved_args
    job_dict["function_kwargs"] = resolved_kwargs
    return job_dict



from __future__ import annotations

import datetime
import time

from monty.json import jsanitize
from rich.scope import render_scope
from rich.table import Table
from rich.text import Text

from jobflow_remote.cli.utils import ReprStr, fmt_datetime
from jobflow_remote.config.base import ExecutionConfig, WorkerBase
from jobflow_remote.jobs.data import FlowInfo, JobDoc, JobInfo
from jobflow_remote.jobs.state import JobState
from jobflow_remote.utils.data import convert_utc_time


def get_job_info_table(jobs_info: list[JobInfo], verbosity: int):
    time_zone_str = f" [{time.tzname[0]}]"

    table = Table(title="Jobs info")
    table.add_column("DB id")
    table.add_column("Name")
    table.add_column("State")
    table.add_column("Job id  (Index)")

    table.add_column("Worker")
    table.add_column("Last updated" + time_zone_str)

    if verbosity >= 1:
        table.add_column("Queue id")
        table.add_column("Run time")
        table.add_column("Retry time" + time_zone_str)
        table.add_column("Prev state")
        if verbosity < 2:
            table.add_column("Locked")

    if verbosity >= 2:
        table.add_column("Lock id")
        table.add_column("Lock time" + time_zone_str)

    for ji in jobs_info:
        state = ji.state.name

        if ji.remote.retry_time_limit is not None:
            state = f"[bold red]{state}[/]"

        row = [
            str(ji.db_id),
            ji.name,
            Text.from_markup(state),
            f"{ji.uuid}  ({ji.index})",
            ji.worker,
            convert_utc_time(ji.updated_on).strftime(fmt_datetime),
        ]

        if verbosity >= 1:
            row.append(ji.remote.process_id)
            prefix = ""
            if ji.state == JobState.RUNNING:
                run_time = ji.estimated_run_time
                prefix = "~"
            else:
                run_time = ji.run_time
            if run_time:
                m, s = divmod(run_time, 60)
                h, m = divmod(m, 60)
                row.append(prefix + f"{h:g}:{m:02g}")
            else:
                row.append("")
            row.append(
                convert_utc_time(ji.remote.retry_time_limit).strftime(fmt_datetime)
                if ji.remote.retry_time_limit
                else None
            )
            row.append(ji.previous_state.name if ji.previous_state else None)
            if verbosity < 2:
                row.append("*" if ji.lock_id is not None else None)

        if verbosity >= 2:
            row.append(str(ji.lock_id))
            row.append(
                convert_utc_time(ji.lock_time).strftime(fmt_datetime)
                if ji.lock_time
                else None
            )

        table.add_row(*row)

    return table


def get_flow_info_table(flows_info: list[FlowInfo], verbosity: int):
    time_zone_str = f" [{time.tzname[0]}]"

    table = Table(title="Flows info")
    table.add_column("DB id")
    table.add_column("Name")
    table.add_column("State")
    table.add_column("Flow id")
    table.add_column("Num Jobs")
    table.add_column("Last updated" + time_zone_str)

    if verbosity >= 1:
        table.add_column("Workers")

        table.add_column("Job states")

    for fi in flows_info:
        # show the smallest Job db_id as db_id
        db_id = min(fi.db_ids)

        row = [
            str(db_id),
            fi.name,
            fi.state.name,
            fi.flow_id,
            str(len(fi.job_ids)),
            convert_utc_time(fi.updated_on).strftime(fmt_datetime),
        ]

        if verbosity >= 1:
            workers = set(fi.workers)
            row.append(", ".join(workers))
            job_states = "-".join(js.short_value for js in fi.job_states)
            row.append(job_states)

        table.add_row(*row)

    return table


def format_job_info(
    job_info: JobInfo | JobDoc, verbosity: int, show_none: bool = False
):
    d = job_info.dict(exclude_none=not show_none)
    if verbosity == 1:
        d.pop("job", None)

    # convert dates at the first level and for the remote error
    for k, v in d.items():
        if isinstance(v, datetime.datetime):
            d[k] = convert_utc_time(v).strftime(fmt_datetime)

    if d["remote"].get("retry_time_limit"):
        d["remote"]["retry_time_limit"] = convert_utc_time(
            d["remote"]["retry_time_limit"]
        ).strftime(fmt_datetime)

    d = jsanitize(d, allow_bson=True, enum_values=True, strict=True)
    error = d.get("error")
    if error:
        d["error"] = ReprStr(error)

    remote_error = d["remote"].get("error")
    if remote_error:
        d["remote"]["error"] = ReprStr(remote_error)

    return render_scope(d)


def format_flow_info(flow_info: FlowInfo):
    title = f"Flow: {flow_info.name} - {flow_info.flow_id} - {flow_info.state.name}"
    table = Table(title=title)
    table.title_style = "bold"
    table.add_column("DB id")
    table.add_column("Name")
    table.add_column("State [Remote]")
    table.add_column("Job id  (Index)")
    table.add_column("Worker")

    for i, job_id in enumerate(flow_info.job_ids):
        state = flow_info.job_states[i].name

        row = [
            str(flow_info.db_ids[i]),
            flow_info.job_names[i],
            state,
            f"{job_id}  ({flow_info.job_indexes[i]})",
            flow_info.workers[i],
        ]

        table.add_row(*row)

    return table


def get_exec_config_table(exec_config: dict[str, ExecutionConfig], verbosity: int = 0):
    table = Table(title="Execution config", show_lines=verbosity > 0)
    table.add_column("Name")
    if verbosity > 0:
        table.add_column("modules")
        table.add_column("export")
        table.add_column("pre_run")
        table.add_column("post_run")
    for name in sorted(exec_config.keys()):
        row = [Text(name, style="bold")]
        if verbosity > 0:
            ec = exec_config[name]
            from ruamel import yaml

            if ec.modules:
                row.append(yaml.dump(ec.modules, default_flow_style=False))
            else:
                row.append("")
            if ec.export:
                row.append(yaml.dump(ec.export, default_flow_style=False))
            else:
                row.append("")
            if ec.post_run:
                row.append(ec.post_run)
            else:
                row.append("")

        table.add_row(*row)

    return table


def get_worker_table(workers: dict[str, WorkerBase], verbosity: int = 0):
    table = Table(title="Workers", show_lines=verbosity > 1)
    table.add_column("Name")
    if verbosity > 0:
        table.add_column("type")
    if verbosity == 1:
        table.add_column("info")
    elif verbosity > 1:
        table.add_column("details")

    for name in sorted(workers.keys()):
        row = [Text(name, style="bold")]
        worker = workers[name]
        if verbosity > 0:
            row.append(worker.type)
        if verbosity == 1:
            row.append(render_scope(worker.cli_info))
        elif verbosity > 1:
            d = worker.dict(exclude_none=True)
            d = jsanitize(d, allow_bson=False, enum_values=True)
            row.append(render_scope(d))

        table.add_row(*row)

    return table



from __future__ import annotations

import logging
import os
import random
from pathlib import Path

from flufl.lock import Lock, LockError

from jobflow_remote.remote.host import BaseHost

logger = logging.getLogger(__name__)


LOCK_DIR = "lock"

TERMINATED_DIR = "terminated"

RUNNING_DIR = "running"

SUBMITTED_DIR = "submitted"


class RemoteBatchManager:
    """
    Manager of remote files containing information about Jobs to be handled by
    a batch worker.

    Used by the Runner.
    """

    def __init__(
        self,
        host: BaseHost,
        files_dir: str | Path,
    ):
        """

        Parameters
        ----------
        host
            The host where the files are.
        files_dir
            The full path to directory where the files are stored.
        """
        self.host = host
        self.files_dir = Path(files_dir)
        self.submitted_dir = self.files_dir / SUBMITTED_DIR
        self.running_dir = self.files_dir / RUNNING_DIR
        self.terminated_dir = self.files_dir / TERMINATED_DIR
        self.lock_dir = self.files_dir / LOCK_DIR
        self._init_files_dir()

    def _init_files_dir(self):
        """
        Initialize the file directory, creating all the subdiretories.
        """
        self.host.connect()
        # Note that the check of the creation of the folders on a remote host
        # slows down the start of the runner by a few seconds.
        # If this proves to be an issue the folder creation should be moved
        # somewhere else and guaranteed in some other way (e.g. a CLI command
        # for the user?).
        self.host.mkdir(self.files_dir)
        self.host.mkdir(self.submitted_dir)
        self.host.mkdir(self.running_dir)
        self.host.mkdir(self.terminated_dir)
        self.host.mkdir(self.lock_dir)

    def submit_job(self, job_id: str, index: int):
        """
        Submit a Job by uploading the corresponding file.

        Parameters
        ----------
        job_id
            Uuid of the Job.
        index
            Index of the Job.
        """
        self.host.write_text_file(self.submitted_dir / f"{job_id}_{index}", "")

    def get_submitted(self) -> list[str]:
        """
        Get a list of files present in the submitted directory.

        Returns
        -------
            The list of file names in the directory.
        """
        return self.host.listdir(self.submitted_dir)

    def get_terminated(self) -> list[tuple[str, int, str]]:
        """
        Get job ids and process ids of the terminated jobs from the corresponding
        directory.

        Returns
        -------

        """
        terminated = []
        for i in self.host.listdir(self.terminated_dir):
            job_id, index, process_uuid = i.split("_")
            index = int(index)
            terminated.append((job_id, index, process_uuid))
        return terminated

    def get_running(self) -> list[tuple[str, int, str]]:
        running = []
        for filename in self.host.listdir(self.running_dir):
            job_id, index, process_uuid = filename.split("_")
            index = int(index)
            running.append((job_id, index, process_uuid))
        return running

    def delete_terminated(self, ids: list[tuple[str, int, str]]):
        for job_id, index, process_uuid in ids:
            self.host.remove(self.terminated_dir / f"{job_id}_{index}_{process_uuid}")


class LocalBatchManager:
    """
    Manager of local files  containing information about Jobs to be handled by
    a batch worker.

    Used in the worker to executes the batch Jobs.
    """

    def __init__(self, files_dir: str | Path, process_id: str):
        self.process_id = process_id
        self.files_dir = Path(files_dir)
        self.submitted_dir = self.files_dir / SUBMITTED_DIR
        self.running_dir = self.files_dir / RUNNING_DIR
        self.terminated_dir = self.files_dir / TERMINATED_DIR
        self.lock_dir = self.files_dir / LOCK_DIR

    def get_job(self) -> str | None:
        files = os.listdir(self.submitted_dir)

        while files:
            selected = random.choice(files)
            try:
                with Lock(
                    str(self.lock_dir / selected), lifetime=60, default_timeout=0
                ):
                    os.remove(self.submitted_dir / selected)
                    (self.running_dir / f"{selected}_{self.process_id}").touch()
                    return selected
            except (LockError, FileNotFoundError):
                logger.error(
                    f"Error while locking file {selected}. Will be ignored",
                    exc_info=True,
                )
                files.remove(selected)
        return None

    def terminate_job(self, job_id: str, index: int):
        os.remove(self.running_dir / f"{job_id}_{index}_{self.process_id}")
        (self.terminated_dir / f"{job_id}_{index}_{self.process_id}").touch()



from typing import Annotated

import typer
from rich.prompt import Confirm
from rich.text import Text

from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.types import (
    db_ids_opt,
    end_date_opt,
    force_opt,
    job_ids_indexes_opt,
    job_state_opt,
    start_date_opt,
)
from jobflow_remote.cli.utils import (
    check_stopped_runner,
    get_config_manager,
    get_job_controller,
    get_job_ids_indexes,
    loading_spinner,
    out_console,
)

app_admin = JFRTyper(
    name="admin", help="Commands for administering the database", no_args_is_help=True
)
app.add_typer(app_admin)


@app_admin.command()
def reset(
    reset_output: Annotated[
        bool,
        typer.Option(
            "--output",
            "-o",
            help="Also delete all the documents in the current store",
        ),
    ] = False,
    max_limit: Annotated[
        int,
        typer.Option(
            "--max",
            "-m",
            help=(
                "The database will be reset only if the number of Flows is lower than the specified limit. 0 means no limit"
            ),
        ),
    ] = 25,
    force: force_opt = False,
):
    """
    Reset the jobflow database.
    WARNING: deletes all the data. These could not be retrieved anymore.
    """
    from jobflow_remote import SETTINGS

    check_stopped_runner(error=True)

    if not force:
        cm = get_config_manager()
        project_name = cm.get_project_data().project.name
        text = Text.from_markup(
            "[red]This operation will [bold]delete all the Jobs data[/bold] "
            f"for project [bold]{project_name}[/bold]. Proceed anyway?[/red]"
        )

        confirmed = Confirm.ask(text, default=False)
        if not confirmed:
            raise typer.Exit(0)
    with loading_spinner(False) as progress:
        progress.add_task(description="Resetting the DB...", total=None)
        jc = get_job_controller()
        done = jc.reset(reset_output=reset_output, max_limit=max_limit)
    not_text = "" if done else "[bold]NOT [/bold]"
    out_console.print(f"The database was {not_text}reset")
    if not done and SETTINGS.cli_suggestions:
        out_console.print(
            "Check the amount of Flows and change --max-limit if this is the correct project to reset",
            style="yellow",
        )


@app_admin.command()
def remove_lock(
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    force: force_opt = False,
):
    """
    Forcibly removes the lock from the documents of the selected jobs.
    WARNING: can lead to inconsistencies if the processes is actually running
    """

    job_ids_indexes = get_job_ids_indexes(job_id)

    jc = get_job_controller()

    if not force:
        with loading_spinner(False) as progress:
            progress.add_task(
                description="Checking the number of locked documents...", total=None
            )

            jobs_info = jc.get_jobs_info(
                job_ids=job_ids_indexes,
                db_ids=db_id,
                state=state,
                start_date=start_date,
                locked=True,
                end_date=end_date,
            )

        text = Text.from_markup(
            f"[red]This operation will [bold]remove the lock[/bold] for (roughly) [bold]{len(jobs_info)} Job(s)[/bold]. Proceed anyway?[/red]"
        )
        confirmed = Confirm.ask(text, default=False)

        if not confirmed:
            raise typer.Exit(0)

        with loading_spinner(False) as progress:
            progress.add_task(
                description="Checking the number of locked documents...", total=None
            )

            num_unlocked = jc.remove_lock_job(
                job_ids=job_id,
                db_ids=db_id,
                state=state,
                start_date=start_date,
                end_date=end_date,
            )

        out_console.print(f"{num_unlocked} jobs were unlocked")



from __future__ import annotations

import abc
import traceback
from pathlib import Path

from monty.json import MSONable


class BaseHost(MSONable):
    """Base Host class."""

    @abc.abstractmethod
    def execute(
        self,
        command: str | list[str],
        workdir: str | Path | None = None,
        timeout: int | None = None,
    ):
        """Execute the given command on the host

        Parameters
        ----------
        command: str or list of str
            Command to execute, as a str or list of str
        workdir: str or None
            path where the command will be executed.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def mkdir(
        self, directory: str | Path, recursive: bool = True, exist_ok: bool = True
    ) -> bool:
        """Create directory on the host."""
        raise NotImplementedError

    @abc.abstractmethod
    def write_text_file(self, filepath, content):
        """Write content to a file on the host."""
        raise NotImplementedError

    @abc.abstractmethod
    def connect(self):
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> bool:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def is_connected(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def put(self, src, dst):
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, src, dst):
        raise NotImplementedError

    @abc.abstractmethod
    def copy(self, src, dst):
        raise NotImplementedError

    def test(self) -> str | None:
        msg = None
        try:
            cmd = "echo 'test'"
            stdout, stderr, returncode = self.execute(cmd)
            if returncode != 0 or stdout.strip() != "test":
                msg = f"Command was executed but some error occurred.\nstdoud: {stdout}\nstderr: {stderr}"
        except Exception:
            exc = traceback.format_exc()
            msg = f"Error while executing command:\n {exc}"

        return msg

    @abc.abstractmethod
    def listdir(self, path: str | Path) -> list[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def remove(self, path: str | Path):
        raise NotImplementedError


class HostError(Exception):
    pass



from __future__ import annotations

import os
from collections.abc import Mapping, MutableMapping
from copy import deepcopy
from datetime import datetime, timezone
from typing import Any
from uuid import UUID

import maggma.stores  # required to enable subclass searching
from maggma.core.store import Store
from monty.json import MontyDecoder


def deep_merge_dict(
    d1: MutableMapping,
    d2: Mapping,
    path: list[str] | None = None,
    raise_on_conflicts: bool = True,
    inplace: bool = True,
) -> MutableMapping:
    """
    Merge a dictionary d2 into a dictionary d1 recursively.

    Parameters
    ----------
    d1
    d2
    path
    raise_on_conflicts
    inplace

    Returns
    -------

    """
    if not inplace:
        d1 = deepcopy(d1)
    if path is None:
        path = []
    for key in d2:
        if key in d1:
            if isinstance(d1[key], Mapping) and isinstance(d2[key], Mapping):
                deep_merge_dict(d1[key], d2[key], path + [str(key)])
            elif d1[key] == d2[key]:
                pass  # same leaf value
            elif raise_on_conflicts:
                raise ValueError("Conflict at %s" % ".".join(path + [str(key)]))
            else:
                d1[key] = d2[key]
        else:
            d1[key] = d2[key]
    return d1


def remove_none(obj):
    if isinstance(obj, (list, tuple, set)):
        return type(obj)(remove_none(x) for x in obj if x is not None)
    elif isinstance(obj, dict):
        return type(obj)(
            (remove_none(k), remove_none(v))
            for k, v in obj.items()
            if k is not None and v is not None
        )
    else:
        return obj


def check_dict_keywords(obj: Any, keywords: list[str]) -> bool:
    if isinstance(obj, (list, tuple, set)):
        return any(check_dict_keywords(x, keywords) for x in obj)
    elif isinstance(obj, dict):
        for k, v in obj.items():
            if isinstance(k, str) and any(k.startswith(kw) for kw in keywords):
                return True
            if check_dict_keywords(v, keywords):
                return True
    return False


def uuid_to_path(
    uuid: str, index: int | None = 1, num_subdirs: int = 3, subdir_len: int = 2
):
    u = UUID(uuid)
    u_hex = u.hex

    # Split the digest into groups of "subdir_len" characters
    subdirs = [
        u_hex[i : i + subdir_len]
        for i in range(0, num_subdirs * subdir_len, subdir_len)
    ]

    # add the index to the final dir name
    dir_name = f"{uuid}"
    if index is not None:
        dir_name += f"_{index}"

    # Combine root directory and subdirectories to form the final path
    return os.path.join(*subdirs, dir_name)


def store_from_dict(store_dict: dict) -> Store:
    if "@class" in store_dict and "@module" in store_dict:
        store = MontyDecoder().process_decoded(store_dict)
        if not isinstance(store, Store):
            raise ValueError(
                f"The converted object {store} is not an instance of a maggma Store"
            )
        return store
    else:

        def all_subclasses(cl):
            return set(cl.__subclasses__()).union(
                [s for c in cl.__subclasses__() for s in all_subclasses(c)]
            )

        all_stores = {s.__name__: s for s in all_subclasses(maggma.stores.Store)}
        return convert_store(store_dict, all_stores)


def convert_store(spec_dict: dict, valid_stores) -> Store:
    """
    Build a store based on the dict spec configuration from JobFlow
    TODO expose the methods from jobflow and don't duplicate the code
    """

    _spec_dict = dict(spec_dict)
    store_type = _spec_dict.pop("type")
    for k, v in _spec_dict.items():
        if isinstance(v, dict) and "type" in v:
            _spec_dict[k] = convert_store(v, valid_stores)
    return valid_stores[store_type](**_spec_dict)


def convert_utc_time(datetime_value: datetime) -> datetime:
    """
    Convert a time in UTC (used in the DB) to the time zone of the
    system where the code is being executed.

    Parameters
    ----------
    datetime_value
        a datetime object in UTC
    Returns
    -------
        The datetime in the zone of the current system
    """
    return datetime_value.replace(tzinfo=timezone.utc).astimezone(tz=None)



from __future__ import annotations

from typing import Callable

from jobflow import Flow, Job, JobStore
from qtoolkit.core.data_objects import QResources

from jobflow_remote.config import ConfigManager
from jobflow_remote.config.base import ExecutionConfig


def set_run_config(
    flow_or_job: Flow | Job,
    name_filter: str = None,
    function_filter: Callable = None,
    exec_config: str | ExecutionConfig | None = None,
    resources: dict | QResources | None = None,
    worker: str | None = None,
) -> Flow | Job:
    """
    Modify in place a Flow or a Job by setting the properties in the
    "manager_config" entry in the JobConfig associated to each Job
    matching the filter. Uses the Flow/Job update_config() method,
    so follows the same conventions, also setting the options in
    the config_updates of the Job, to allow setting the same properties
    also in dynamically generated Jobs.

    Parameters
    ----------
    flow_or_job
        A Flow or a Job to be modified
    name_filter
        A filter for the job name. Only jobs with a matching name will be updated.
        Includes partial matches, e.g. "ad" will match a job with the name "adder".
    function_filter
        A filter for the job function. Only jobs with a matching function will be
        updated.
    exec_config
        The execution configuration to be added to the selected Jobs.
    resources
        The resources to be set for the selected Jobs.
    worker
        The worker where the selected Jobs will be executed.

    Returns
    -------
    Flow or Job
        The modified object.
    """
    if not exec_config and not resources and not worker:
        return flow_or_job
    config: dict = {"manager_config": {}}
    if exec_config:
        config["manager_config"]["exec_config"] = exec_config
    if resources:
        config["manager_config"]["resources"] = resources
    if worker:
        config["manager_config"]["worker"] = worker

    flow_or_job.update_config(
        config=config, name_filter=name_filter, function_filter=function_filter
    )

    return flow_or_job


def load_job_store(project: str | None = None) -> JobStore:
    """
    Load the JobStore for the current project.

    Parameters
    ----------
    project

    Returns
    -------

    """
    cm = ConfigManager()
    p = cm.get_project(project)
    job_store = p.get_jobstore()

    return job_store



import io
from pathlib import Path
from typing import Annotated, Optional

import typer
from monty.json import jsanitize
from monty.serialization import dumpfn
from qtoolkit.core.data_objects import QResources
from rich.pretty import pprint

from jobflow_remote import SETTINGS
from jobflow_remote.cli.formatting import format_job_info, get_job_info_table
from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.types import (
    break_lock_opt,
    days_opt,
    db_ids_opt,
    end_date_opt,
    flow_ids_opt,
    hours_opt,
    job_db_id_arg,
    job_ids_indexes_opt,
    job_index_arg,
    job_state_arg,
    job_state_opt,
    locked_opt,
    max_results_opt,
    metadata_opt,
    name_opt,
    query_opt,
    raise_on_error_opt,
    reverse_sort_flag_opt,
    sort_opt,
    start_date_opt,
    verbosity_opt,
    wait_lock_opt,
)
from jobflow_remote.cli.utils import (
    SortOption,
    check_incompatible_opt,
    check_stopped_runner,
    execute_multi_jobs_cmd,
    exit_with_error_msg,
    get_config_manager,
    get_job_controller,
    get_job_db_ids,
    get_job_ids_indexes,
    get_start_date,
    loading_spinner,
    out_console,
    print_success_msg,
    str_to_dict,
)
from jobflow_remote.remote.queue import ERR_FNAME, OUT_FNAME

app_job = JFRTyper(
    name="job", help="Commands for managing the jobs", no_args_is_help=True
)
app.add_typer(app_job)


@app_job.command(name="list")
def jobs_list(
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    max_results: max_results_opt = 100,
    sort: sort_opt = SortOption.UPDATED_ON.value,
    reverse_sort: reverse_sort_flag_opt = False,
    locked: locked_opt = False,
    custom_query: query_opt = None,
):
    """
    Get the list of Jobs in the database
    """
    check_incompatible_opt({"start_date": start_date, "days": days, "hours": hours})
    check_incompatible_opt({"end_date": end_date, "days": days, "hours": hours})
    metadata_dict = str_to_dict(metadata)

    job_ids_indexes = get_job_ids_indexes(job_id)

    jc = get_job_controller()

    start_date = get_start_date(start_date, days, hours)

    sort = [(sort.value, 1 if reverse_sort else -1)]

    with loading_spinner():
        if custom_query:
            jobs_info = jc.get_jobs_info_query(
                query=custom_query,
                limit=max_results,
                sort=sort,
            )
        else:
            jobs_info = jc.get_jobs_info(
                job_ids=job_ids_indexes,
                db_ids=db_id,
                flow_ids=flow_id,
                state=state,
                start_date=start_date,
                locked=locked,
                end_date=end_date,
                name=name,
                metadata=metadata_dict,
                limit=max_results,
                sort=sort,
            )

        table = get_job_info_table(jobs_info, verbosity=verbosity)

    out_console.print(table)
    if SETTINGS.cli_suggestions:
        if max_results and len(jobs_info) == max_results:
            out_console.print(
                f"The number of Jobs printed may be limited by the maximum selected: {max_results}",
                style="yellow",
            )
        if any(ji.remote.retry_time_limit is not None for ji in jobs_info):
            text = (
                "Some jobs (state in red) have failed while interacting with"
                " the worker, but will be retried again.\nGet more information about"
                " the error with 'jf job info JOB_ID'"
            )
            out_console.print(text, style="yellow")


@app_job.command(name="info")
def job_info(
    job_db_id: job_db_id_arg,
    job_index: job_index_arg = None,
    with_error: Annotated[
        bool,
        typer.Option(
            "--with-error",
            "-err",
            help="DEPRECATED: not needed anymore to fetch errors",
        ),
    ] = False,
    show_none: Annotated[
        bool,
        typer.Option(
            "--show-none",
            "-n",
            help="Show the data whose values are None. Usually hidden",
        ),
    ] = False,
    verbosity: verbosity_opt = 0,
):
    """
    Detail information on a specific job
    """

    db_id, job_id = get_job_db_ids(job_db_id, job_index)

    if with_error:
        out_console.print(
            "The --with-error option is deprecated and not needed anymore to show error messages",
            style="yellow",
        )

    with loading_spinner():
        jc = get_job_controller()

        if verbosity > 0:
            job_data = jc.get_job_doc(
                job_id=job_id,
                job_index=job_index,
                db_id=db_id,
            )
        else:
            job_data = jc.get_job_info(
                job_id=job_id,
                job_index=job_index,
                db_id=db_id,
            )

    if not job_data:
        exit_with_error_msg("No data matching the request")

    out_console.print(format_job_info(job_data, verbosity, show_none=show_none))


@app_job.command()
def set_state(
    state: job_state_arg,
    job_db_id: job_db_id_arg,
    job_index: job_index_arg = None,
):
    """
    Sets the state of a Job to an arbitrary value.
    WARNING: No checks. This can lead to inconsistencies in the DB. Use with care
    """

    db_id, job_id = get_job_db_ids(job_db_id, job_index)

    with loading_spinner():
        jc = get_job_controller()

        succeeded = jc.set_job_state(
            state=state,
            job_id=job_id,
            job_index=job_index,
            db_id=db_id,
        )

    if not succeeded:
        exit_with_error_msg("Could not reset the remote attempts")

    print_success_msg()


@app_job.command()
def rerun(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    break_lock: break_lock_opt = False,
    force: Annotated[
        bool,
        typer.Option(
            "--force",
            "-f",
            help=(
                "Force the rerun even if some conditions would normally prevent it (e.g. "
                "state usually not allowed or children already executed). Can lead to "
                "inconsistencies. Advanced users."
            ),
        ),
    ] = False,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Rerun a Job. By default, this is limited to jobs that failed and children did
    not start or jobs that are running. The rerun Job is set to READY and children
    Jobs to WAITING. If possible, the associated job submitted to the remote queue
    will be cancelled. Most of the limitations can be overridden by the 'force'
    option. This could lead to inconsistencies in the overall state of the Jobs of
    the Flow.
    """
    if force or break_lock:
        check_stopped_runner(error=False)

    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.rerun_job,
        multi_cmd=jc.rerun_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        state=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        verbosity=verbosity,
        wait=wait,
        break_lock=break_lock,
        force=force,
        raise_on_error=raise_on_error,
    )


@app_job.command()
def retry(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    break_lock: break_lock_opt = False,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Retry to perform the operation that failed for a job in a REMOTE_ERROR state
    or reset the number of attempts at remote action, in order to allow the
    runner to try it again immediately.
    """
    if break_lock:
        check_stopped_runner(error=False)

    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.retry_job,
        multi_cmd=jc.retry_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        state=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        verbosity=verbosity,
        wait=wait,
        break_lock=break_lock,
        raise_on_error=raise_on_error,
    )


@app_job.command()
def pause(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Pause a Job. Only READY and WAITING Jobs can be paused. The operation is reversible.
    """

    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.pause_job,
        multi_cmd=jc.pause_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        state=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        verbosity=verbosity,
        wait=wait,
        raise_on_error=raise_on_error,
    )


@app_job.command()
def play(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Resume a Job that was previously PAUSED.
    """

    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.play_job,
        multi_cmd=jc.play_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        state=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        verbosity=verbosity,
        wait=wait,
        raise_on_error=raise_on_error,
    )


@app_job.command()
def stop(
    job_db_id: job_db_id_arg = None,
    job_index: job_index_arg = None,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    wait: wait_lock_opt = None,
    break_lock: break_lock_opt = False,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Stop a Job. Only Jobs that did not complete or had an error can be stopped.
    The operation is irreversible.
    If possible, the associated job submitted to the remote queue will be cancelled.
    """
    if break_lock:
        check_stopped_runner(error=False)

    jc = get_job_controller()

    execute_multi_jobs_cmd(
        single_cmd=jc.stop_job,
        multi_cmd=jc.stop_jobs,
        job_db_id=job_db_id,
        job_index=job_index,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        state=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        verbosity=verbosity,
        wait=wait,
        break_lock=break_lock,
        raise_on_error=raise_on_error,
    )


@app_job.command()
def queue_out(
    job_db_id: job_db_id_arg,
    job_index: job_index_arg = None,
):
    """
    Print the content of the output files produced by the queue manager.
    """

    db_id, job_id = get_job_db_ids(job_db_id, job_index)

    cm = get_config_manager()

    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Retrieving info...", total=None)
        jc = get_job_controller()

        job_info = jc.get_job_info(
            job_id=job_id,
            job_index=job_index,
            db_id=db_id,
        )

    if not job_info:
        exit_with_error_msg("No data matching the request")

    remote_dir = job_info.run_dir

    out_path = Path(remote_dir, OUT_FNAME)
    err_path = Path(remote_dir, ERR_FNAME)
    out = None
    err = None
    out_error = None
    err_error = None
    with loading_spinner(processing=False) as progress:
        progress.add_task(description="Retrieving files...", total=None)
        worker = cm.get_worker(job_info.worker)
        host = worker.get_host()

        try:
            host.connect()
            try:
                out_bytes = io.BytesIO()
                host.get(out_path, out_bytes)
                out = out_bytes.getvalue().decode("utf-8")
            except Exception as e:
                out_error = getattr(e, "message", str(e))
            try:
                err_bytes = io.BytesIO()
                host.get(err_path, err_bytes)
                err = err_bytes.getvalue().decode("utf-8")
            except Exception as e:
                err_error = getattr(e, "message", str(e))
        finally:
            try:
                host.close()
            except Exception:
                pass

    if out_error:
        out_console.print(
            f"Error while fetching queue output from {str(out_path)}: {out_error}",
            style="red",
        )
    else:
        out_console.print(f"Queue output from {str(out_path)}:\n")
        out_console.print(out)

    if err_error:
        out_console.print(
            f"Error while fetching queue error from {str(err_path)}: {err_error}",
            style="red",
        )
    else:
        out_console.print(f"Queue error from {str(err_path)}:\n")
        out_console.print(err)


app_job_set = JFRTyper(
    name="set", help="Commands for managing the jobs", no_args_is_help=True
)
app_job.add_typer(app_job_set)


@app_job_set.command()
def worker(
    worker_name: Annotated[
        str,
        typer.Argument(
            help="The name of the worker",
            metavar="WORKER",
        ),
    ],
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Set the worker for the selected Jobs. Only READY or WAITING Jobs.
    """

    jc = get_job_controller()
    execute_multi_jobs_cmd(
        single_cmd=jc.set_job_run_properties,
        multi_cmd=jc.set_job_run_properties,
        job_db_id=None,
        job_index=None,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        state=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        verbosity=verbosity,
        raise_on_error=raise_on_error,
        worker=worker_name,
    )


@app_job_set.command()
def exec_config(
    exec_config_value: Annotated[
        str,
        typer.Argument(
            help="The name of the exec_config",
            metavar="EXEC_CONFIG",
        ),
    ],
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Set the exec_config for the selected Jobs. Only READY or WAITING Jobs.
    """

    jc = get_job_controller()
    execute_multi_jobs_cmd(
        single_cmd=jc.set_job_run_properties,
        multi_cmd=jc.set_job_run_properties,
        job_db_id=None,
        job_index=None,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        state=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        verbosity=verbosity,
        raise_on_error=raise_on_error,
        exec_config_value=exec_config_value,
    )


@app_job_set.command()
def resources(
    resources_value: Annotated[
        str,
        typer.Argument(
            help="The resources to be specified. Can be either a list of"
            "comma separated key=value pairs or a string with the JSON "
            "representation of a dictionary "
            '(e.g \'{"key1.key2": 1, "key3": "test"}\')',
            metavar="EXEC_CONFIG",
        ),
    ],
    replace: Annotated[
        bool,
        typer.Option(
            "--replace",
            "-r",
            help="If present the value will replace entirely those present "
            "instead of updating the DB, otherwise only the selected keys "
            "will be updated",
        ),
    ] = False,
    qresources: Annotated[
        bool,
        typer.Option(
            "--qresources",
            "-qr",
            help="If present the values will be interpreted as arguments for a QResources object",
        ),
    ] = False,
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    raise_on_error: raise_on_error_opt = False,
):
    """
    Set the resources for the selected Jobs. Only READY or WAITING Jobs.
    """

    resources_value = str_to_dict(resources_value)

    if qresources:
        resources_value = QResources(**resources_value)

    jc = get_job_controller()
    execute_multi_jobs_cmd(
        single_cmd=jc.set_job_run_properties,
        multi_cmd=jc.set_job_run_properties,
        job_db_id=None,
        job_index=None,
        job_ids=job_id,
        db_ids=db_id,
        flow_ids=flow_id,
        state=state,
        start_date=start_date,
        end_date=end_date,
        name=name,
        metadata=metadata,
        days=days,
        hours=hours,
        verbosity=verbosity,
        raise_on_error=raise_on_error,
        resources=resources_value,
        update=not replace,
    )


@app_job.command(name="dump", hidden=True)
def job_dump(
    job_id: job_ids_indexes_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: job_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    metadata: metadata_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    file_path: Annotated[
        str,
        typer.Option(
            "--path",
            "-p",
            help="Path to where the file should be dumped",
        ),
    ] = "jobs_dump.json",
):
    """
    Dump to json the documents of the selected Jobs from the DB. For debugging.
    """

    check_incompatible_opt({"start_date": start_date, "days": days, "hours": hours})
    check_incompatible_opt({"end_date": end_date, "days": days, "hours": hours})
    metadata_dict = str_to_dict(metadata)

    job_ids_indexes = get_job_ids_indexes(job_id)

    jc = get_job_controller()

    start_date = get_start_date(start_date, days, hours)

    with loading_spinner():
        jobs_doc = jc.get_jobs_doc(
            job_ids=job_ids_indexes,
            db_ids=db_id,
            flow_ids=flow_id,
            state=state,
            start_date=start_date,
            end_date=end_date,
            name=name,
            metadata=metadata_dict,
        )
        if jobs_doc:
            dumpfn(jsanitize(jobs_doc, strict=True, enum_values=True), file_path)

    if not jobs_doc:
        exit_with_error_msg("No data matching the request")


@app_job.command()
def output(
    job_db_id: job_db_id_arg,
    job_index: job_index_arg = None,
    file_path: Annotated[
        Optional[str],
        typer.Option(
            "--path",
            "-p",
            help="If defined, the output will be dumped to this file based on the extension (json or yaml)",
        ),
    ] = None,
    load: Annotated[
        bool,
        typer.Option(
            "--load",
            "-",
            help="If enabled all the data from additional stores are also loaded ",
        ),
    ] = False,
):
    """
    Detail information on a specific job
    """

    db_id, job_id = get_job_db_ids(job_db_id, job_index)

    with loading_spinner():
        jc = get_job_controller()

        if db_id:
            job_info = jc.get_job_info(
                job_id=job_id,
                job_index=job_index,
                db_id=db_id,
            )
            if job_info:
                job_id = job_info.uuid
                job_index = job_info.index

        job_output = None
        if job_id:
            job_output = jc.jobstore.get_output(job_id, job_index or "last", load=load)

    if not job_output:
        exit_with_error_msg("No data matching the request")

    if file_path:
        dumpfn(job_output, file_path)
    else:
        pprint(job_output)



from __future__ import annotations

import copy
import logging
import time
import warnings
from collections import defaultdict
from collections.abc import Iterable, Mapping
from datetime import datetime
from typing import TYPE_CHECKING, Any

from jobflow.utils import suuid
from pymongo import ReturnDocument

from jobflow_remote.utils.data import deep_merge_dict

if TYPE_CHECKING:
    from pymongo.collection import Collection

logger = logging.getLogger(__name__)


class MongoLock:
    """
    Context manager to lock a document in a MongoDB database.

    Main characteristics and functionalities:
        * Lock is acquired by setting a lock_id and lock_time value in the
          locked document.
        * Filter the document to select based on a query and sorting.
          It uses find_one_and_update, thus resulting in a single document locked.
        * Can wait for a lock to be released.
        * Can forcibly break an existing lock
        * Can return the locked document even if the lock could not be acquired
          (useful for determining if a document is locked or no document matches
          the query)
        * Accepts all the arguments that can be passed to find_one_and_update
        * A custom update can be performed on the document when acquiring the lock.
        * Allows to pass properties that will be set in the document at the moment
          of releasing the lock.
        * Lock id value can be customized. If not a randomly generated uuid is used.

    Examples
    --------

    Trying to acquire the lock on a document based on the state
    >>> with MongoLock(collection, {"state": "READY"}) as lock:
    ...     print(lock.locked_document["state"])
    READY

    If lock cannot be acquired (no document matching filter or that
    document is locked) the `lock.locked_document` is None.

    >>> with MongoLock(collection, {"state": "READY"}) as lock:
    ...     print(lock.locked_document)
    None

    Wait for 60 seconds in case the required document is already locked.
    Check the status every 10 seconds. If lock cannot be acquired
    locked_document is None

    >>> with MongoLock(
            collection,
            {"uuid": "5b84228b-d019-47fe-b0a0-564b36aa85ed"},
            sleep=10,
            max_wait=60,
        ) as lock:
    ...     print(lock.locked_document)
    None

    In case lock cannot be acquired expose the locked document

    >>> with MongoLock(
            collection,
            {"uuid": "5b84228b-d019-47fe-b0a0-564b36aa85ed"},
            get_locked_doc=True,
        ) as lock:
    ...     print(lock.locked_document)
    None
    ...     print(lock.unavailable_document['lock_id'])
    8d68404f-c77a-461b-859c-40bb0af1979f

    Set values in the document upon lock release

    >>> with MongoLock(collection, {"state": "READY"}) as lock:
    ...     if lock.locked_document:
    ...         # Perform some operations based on the job...
    ...         lock.update_on_release = {"$set": {"state": "CHECKED_OUT"}}

    """

    LOCK_KEY = "lock_id"
    LOCK_TIME_KEY = "lock_time"

    def __init__(
        self,
        collection: Collection,
        filter: Mapping[str, Any],
        update: Mapping[str, Any] | None = None,
        break_lock: bool = False,
        lock_id: str | None = None,
        sleep: int | None = None,
        max_wait: int = 600,
        projection: Mapping[str, Any] | Iterable[str] | None = None,
        get_locked_doc: bool = False,
        **kwargs,
    ):
        """
        Parameters
        ----------
        collection
            The MongoDB collection containing the document to lock.
        filter
            A MongoDB query to select the document.
        update
            A dictionary that will be passed to find_one_and_update to update the
            locked document at the moment of acquiring the lock.
        break_lock
            True if the context manager is allowed to forcibly break a lock.
        lock_id
            The is used for the lock in the document. If None a randomly generated
            uuid will be used.
        sleep
            The amount of second to sleep between consecutive checks while waiting
            for a lock to be released.
        max_wait
            The amount of seconds to wait for a lock to be released.
        projection
            The projection passed to the find_one_and_update that locks the document.
        get_locked_doc
            If True, if the lock cannot be acquired because the document matching
            the filter is already locked, the locked document will be fetched and
            set in the unavailable_document attribute.
        kwargs
            All the other args are passed to find_one_and_update.
        """
        self.collection = collection
        self.filter = filter or {}
        self.update = update
        self.break_lock = break_lock
        self.locked_document = None
        self.unavailable_document = None
        self.lock_id = lock_id or suuid()
        self.kwargs = kwargs
        self.update_on_release: dict = {}
        self.sleep = sleep
        self.max_wait = max_wait
        self.projection = projection
        self.get_locked_doc = get_locked_doc

    @classmethod
    def get_lock_time(cls, d: dict):
        """
        Get the time the document was locked on a dictionary.
        """
        return d.get(cls.LOCK_TIME_KEY)

    @classmethod
    def get_lock_id(cls, d: dict):
        """
        Get the lock id on a dictionary.
        """
        return d.get(cls.LOCK_KEY)

    def acquire(self):
        """
        Acquire the lock
        """
        # Set the lock expiration time
        now = datetime.utcnow()
        db_filter = copy.deepcopy(self.filter)

        projection = self.projection
        # if projecting always get the lock as well
        if projection:
            projection = list(projection)
            projection.extend([self.LOCK_KEY, self.lock_id])

        # Modify the filter if the document should not be fetched if
        # the lock cannot be acquired. Otherwise, keep the original filter.
        if not self.break_lock and not self.sleep and not self.get_locked_doc:
            db_filter.update({self.LOCK_KEY: None})

        # Prepare the update to be performed when acquiring the lock.
        # A combination of the input update and the setting of the lock.
        lock_set = {self.LOCK_KEY: self.lock_id, self.LOCK_TIME_KEY: now}
        update = defaultdict(dict)
        if self.update:
            update.update(copy.deepcopy(self.update))

        update["$set"].update(lock_set)

        # If the document should be fetched even if the lock could not be acquired
        # the updates should be made conditional.
        # Note: sleep needs to fetch the document, otherwise it is impossible to
        # determine if the filter did not return any document or if the document
        # was locked.
        if (self.sleep or self.get_locked_doc) and not self.break_lock:
            for operation, dict_vals in update.items():
                for k, v in dict_vals.items():
                    cond = {
                        "$cond": {
                            "if": {"$gt": [f"${self.LOCK_KEY}", None]},
                            "then": f"${k}",
                            "else": v,
                        }
                    }
                    update[operation][k] = cond
            update = [dict(update)]

        # Try to acquire the lock by updating the document with a unique identifier
        # and the lock expiration time
        logger.debug(f"try acquiring lock with filter: {db_filter}")
        t0 = time.time()
        while True:
            result = self.collection.find_one_and_update(
                db_filter,
                update,
                upsert=False,
                return_document=ReturnDocument.AFTER,
                **self.kwargs,
            )

            if result:
                lock_acquired = self.get_lock_id(result) == self.lock_id
                if lock_acquired:
                    self.locked_document = result
                    break
                else:
                    # if the lock could not be acquired optionally sleep or
                    # exit if waited for enough time.
                    if self.sleep and (time.time() - t0) < self.max_wait:
                        logger.debug("sleeping")
                        time.sleep(self.sleep)
                    else:
                        self.unavailable_document = result
                        break
            else:
                # If no document the conditions could not be met.
                # Either the requested filter does not find match a document
                # or those fitting are locked.
                break

    def release(self, exc_type, exc_val, exc_tb):
        """
        Release the lock.
        """
        # Release the lock by removing the unique identifier and lock expiration time
        update = {"$set": {self.LOCK_KEY: None, self.LOCK_TIME_KEY: None}}
        # TODO maybe set on release only if no exception was raised?
        if self.update_on_release:
            update = deep_merge_dict(update, self.update_on_release)
        logger.debug(f"release lock with update: {update}")
        # TODO if failed to release the lock maybe retry before failing
        result = self.collection.update_one(
            {"_id": self.locked_document["_id"], self.LOCK_KEY: self.lock_id},
            update,
            upsert=False,
        )

        # Check if the lock was successfully released
        if result.modified_count == 0:
            msg = f"Could not release lock for document {self.locked_document['_id']}"
            warnings.warn(msg, stacklevel=2)

        self.locked_document = None

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.locked_document:
            self.release(exc_type, exc_val, exc_tb)


class LockedDocumentError(Exception):
    """
    Exception to signal a problem when locking the document.
    """


class JobLockedError(LockedDocumentError):
    """
    Exception to signal a problem when locking a Job document.
    """

    @classmethod
    def from_job_doc(cls, doc: dict, additional_msg: str | None = None):
        lock_id = doc[MongoLock.LOCK_KEY]
        lock_date = doc[MongoLock.LOCK_TIME_KEY]
        date_str = lock_date.isoformat(timespec="seconds") if lock_date else None
        msg = f"Job with db_id {doc['db_id']} is locked with lock_id {lock_id} since {date_str} UTC."
        if additional_msg:
            msg += " " + additional_msg
        return cls(msg)


class FlowLockedError(LockedDocumentError):
    """
    Exception to signal a problem when locking a Flow document.
    """

    @classmethod
    def from_flow_doc(cls, doc: dict, additional_msg: str | None = None):
        lock_id = doc[MongoLock.LOCK_KEY]
        lock_date = doc[MongoLock.LOCK_TIME_KEY]
        date_str = lock_date.isoformat(timespec="seconds") if lock_date else None
        msg = f"Flow with uuid {doc['uuid']} is locked with lock_id {lock_id} since {date_str} UTC."
        if additional_msg:
            msg += " " + additional_msg
        return cls(msg)



import os
from typing import Annotated

import typer
from rich.table import Table
from rich.text import Text

from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.types import log_level_opt
from jobflow_remote.cli.utils import (
    exit_with_error_msg,
    exit_with_warning_msg,
    get_config_manager,
    loading_spinner,
    out_console,
)
from jobflow_remote.config.base import LogLevel
from jobflow_remote.jobs.daemon import DaemonError, DaemonManager, DaemonStatus
from jobflow_remote.jobs.runner import Runner

app_runner = JFRTyper(
    name="runner", help="Commands for handling the Runner", no_args_is_help=True
)
app.add_typer(app_runner)


@app_runner.command()
def run(
    log_level: log_level_opt = LogLevel.INFO.value,
    set_pid: Annotated[
        bool,
        typer.Option(
            "--set-pid",
            "-pid",
            help="Set the runner id to the current process pid",
        ),
    ] = False,
    transfer: Annotated[
        bool,
        typer.Option(
            "--transfer",
            "-t",
            help="Enable the transfer option in the runner",
        ),
    ] = False,
    complete: Annotated[
        bool,
        typer.Option(
            "--complete",
            "-com",
            help="Enable the complete option in the runner",
        ),
    ] = False,
    queue: Annotated[
        bool,
        typer.Option(
            "--queue",
            "-q",
            help="Enable the queue option in the runner",
        ),
    ] = False,
    checkout: Annotated[
        bool,
        typer.Option(
            "--checkout",
            "-cho",
            help="Enable the checkout option in the runner",
        ),
    ] = False,
):
    """
    Execute the Runner in the foreground.
    Do NOT execute this to start as a daemon.
    Should be used by the daemon or for testing purposes.
    """
    runner_id = os.getpid() if set_pid else None
    runner = Runner(log_level=log_level, runner_id=str(runner_id))
    if not (transfer or complete or queue or checkout):
        transfer = complete = queue = checkout = True

    runner.run(transfer=transfer, complete=complete, queue=queue, checkout=checkout)


@app_runner.command()
def start(
    transfer: Annotated[
        int,
        typer.Option(
            "--transfer",
            "-t",
            help="The number of processes dedicated to completing jobs",
        ),
    ] = 1,
    complete: Annotated[
        int,
        typer.Option(
            "--complete",
            "-com",
            help="The number of processes dedicated to completing jobs",
        ),
    ] = 1,
    single: Annotated[
        bool,
        typer.Option(
            "--single",
            "-s",
            help="Use a single process for the runner",
        ),
    ] = False,
    log_level: log_level_opt = LogLevel.INFO.value,
):
    """
    Start the Runner as a daemon
    """
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    with loading_spinner(False) as progress:
        progress.add_task(description="Starting the daemon...", total=None)
        try:
            dm.start(
                num_procs_transfer=transfer,
                num_procs_complete=complete,
                single=single,
                log_level=log_level.value,
                raise_on_error=True,
            )
        except DaemonError as e:
            exit_with_error_msg(
                f"Error while starting the daemon: {getattr(e, 'message', e)}"
            )


@app_runner.command()
def stop(
    wait: Annotated[
        bool,
        typer.Option(
            "--wait",
            "-w",
            help=(
                "Wait until the daemon has stopped. NOTE: this may take a while if a large file is being transferred!"
            ),
        ),
    ] = False
):
    """
    Send a stop signal to the Runner processes.
    Each of the Runner processes will stop when finished the task being executed.
    By default, return immediately
    """
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    with loading_spinner(False) as progress:
        progress.add_task(description="Stopping the daemon...", total=None)
        try:
            dm.stop(wait=wait, raise_on_error=True)
        except DaemonError as e:
            exit_with_error_msg(
                f"Error while stopping the daemon: {getattr(e, 'message', e)}"
            )
    from jobflow_remote import SETTINGS

    if SETTINGS.cli_suggestions:
        out_console.print(
            "The stop signal has been sent to the Runner. Run 'jf runner status' to verify if it stopped",
            style="yellow",
        )


@app_runner.command()
def kill():
    """
    Send a kill signal to the Runner processes.
    Return immediately, does not wait for processes to be killed.
    """
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    with loading_spinner(False) as progress:
        progress.add_task(description="Killing the daemon...", total=None)
        try:
            dm.kill(raise_on_error=True)
        except DaemonError as e:
            exit_with_error_msg(
                f"Error while killing the daemon: {getattr(e, 'message', e)}"
            )


@app_runner.command()
def shutdown():
    """
    Shuts down the supervisord process.
    Note that if the daemon is running it will wait for the daemon to stop.
    """
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    with loading_spinner(False) as progress:
        progress.add_task(description="Shutting down supervisor...", total=None)
        try:
            dm.shut_down(raise_on_error=True)
        except DaemonError as e:
            exit_with_error_msg(
                f"Error while shutting down supervisor: {getattr(e, 'message', e)}"
            )


@app_runner.command()
def status():
    """
    Fetch the status of the daemon runner
    """
    from jobflow_remote import SETTINGS

    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    with loading_spinner():
        try:
            current_status = dm.check_status()
        except DaemonError as e:
            exit_with_error_msg(
                f"Error while checking the status of the daemon: {getattr(e, 'message', e)}"
            )
    color = {
        DaemonStatus.STOPPED: "red",
        DaemonStatus.STOPPING: "gold1",
        DaemonStatus.SHUT_DOWN: "red",
        DaemonStatus.PARTIALLY_RUNNING: "gold1",
        DaemonStatus.RUNNING: "green",
    }[current_status]
    text = Text()
    text.append("Daemon status: ")
    text.append(current_status.value.lower(), style=color)
    out_console.print(text)
    if current_status == DaemonStatus.PARTIALLY_RUNNING and SETTINGS.cli_suggestions:
        out_console.print(
            f"The {current_status.value.lower()} may be present due to the "
            "runner stopping or signal a problem with one of the processes "
            "of the runner. If the state should be RUNNING, check the detailed"
            " status with the 'info'  command and consider restarting the runner.",
            style="yellow",
        )


@app_runner.command()
def info():
    """
    Fetch the information about the process of the daemon.
    Contain the supervisord process and the processes running the Runner.
    """
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    procs_info_dict = None
    try:
        with loading_spinner():
            procs_info_dict = dm.get_processes_info()
    except DaemonError as e:
        exit_with_error_msg(
            f"Error while stopping the daemon: {getattr(e, 'message', e)}"
        )
    if not procs_info_dict:
        exit_with_warning_msg("Daemon is not running")
    table = Table()
    table.add_column("Process")
    table.add_column("PID")
    table.add_column("State")

    for name, proc_info in procs_info_dict.items():
        table.add_row(name, str(proc_info["pid"]), str(proc_info["state"]))

    out_console.print(table)



from typing import Annotated, Optional

import typer
from jobflow.utils.graph import draw_graph
from rich.prompt import Confirm
from rich.text import Text

from jobflow_remote import SETTINGS
from jobflow_remote.cli.formatting import format_flow_info, get_flow_info_table
from jobflow_remote.cli.jf import app
from jobflow_remote.cli.jfr_typer import JFRTyper
from jobflow_remote.cli.types import (
    days_opt,
    db_ids_opt,
    end_date_opt,
    flow_db_id_arg,
    flow_ids_opt,
    flow_state_opt,
    force_opt,
    hours_opt,
    job_flow_id_flag_opt,
    job_ids_opt,
    max_results_opt,
    name_opt,
    reverse_sort_flag_opt,
    sort_opt,
    start_date_opt,
    verbosity_opt,
)
from jobflow_remote.cli.utils import (
    SortOption,
    check_incompatible_opt,
    exit_with_error_msg,
    exit_with_warning_msg,
    get_job_controller,
    get_job_db_ids,
    get_start_date,
    loading_spinner,
    out_console,
)
from jobflow_remote.jobs.graph import get_graph, plot_dash

app_flow = JFRTyper(
    name="flow", help="Commands for managing the flows", no_args_is_help=True
)
app.add_typer(app_flow)


@app_flow.command(name="list")
def flows_list(
    job_id: job_ids_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: flow_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    verbosity: verbosity_opt = 0,
    max_results: max_results_opt = 100,
    sort: sort_opt = SortOption.UPDATED_ON.value,
    reverse_sort: reverse_sort_flag_opt = False,
):
    """
    Get the list of Jobs in the database
    """
    check_incompatible_opt({"start_date": start_date, "days": days, "hours": hours})
    check_incompatible_opt({"end_date": end_date, "days": days, "hours": hours})

    jc = get_job_controller()

    start_date = get_start_date(start_date, days, hours)

    sort = [(sort.value, 1 if reverse_sort else -1)]

    with loading_spinner():
        flows_info = jc.get_flows_info(
            job_ids=job_id,
            db_ids=db_id,
            flow_ids=flow_id,
            state=state,
            start_date=start_date,
            end_date=end_date,
            name=name,
            limit=max_results,
            sort=sort,
            full=verbosity > 0,
        )

        table = get_flow_info_table(flows_info, verbosity=verbosity)

    if SETTINGS.cli_suggestions:
        if max_results and len(flows_info) == max_results:
            out_console.print(
                f"The number of Flows printed is limited by the maximum selected: {max_results}",
                style="yellow",
            )

    out_console.print(table)


@app_flow.command()
def delete(
    job_id: job_ids_opt = None,
    db_id: db_ids_opt = None,
    flow_id: flow_ids_opt = None,
    state: flow_state_opt = None,
    start_date: start_date_opt = None,
    end_date: end_date_opt = None,
    name: name_opt = None,
    days: days_opt = None,
    hours: hours_opt = None,
    force: force_opt = False,
):
    """
    Permanently delete Flows from the database
    """
    check_incompatible_opt({"start_date": start_date, "days": days, "hours": hours})
    check_incompatible_opt({"end_date": end_date, "days": days, "hours": hours})

    start_date = get_start_date(start_date, days, hours)

    jc = get_job_controller()

    with loading_spinner(False) as progress:
        progress.add_task(description="Fetching data...", total=None)
        flows_info = jc.get_flows_info(
            job_ids=job_id,
            db_ids=db_id,
            flow_ids=flow_id,
            state=state,
            start_date=start_date,
            end_date=end_date,
            name=name,
        )

    if not flows_info:
        exit_with_warning_msg("No flows matching criteria")

    if flows_info and not force:
        text = Text.from_markup(
            f"[red]This operation will [bold]delete {len(flows_info)} Flow(s)[/bold]. Proceed anyway?[/red]"
        )

        confirmed = Confirm.ask(text, default=False)
        if not confirmed:
            raise typer.Exit(0)

    to_delete = [fi.flow_id for fi in flows_info]
    with loading_spinner(False) as progress:
        progress.add_task(description="Deleting...", total=None)

        jc.delete_flows(flow_ids=to_delete)

    out_console.print(
        f"Deleted Flow(s) with id: {', '.join(str(i) for i in to_delete)}"
    )


@app_flow.command(name="info")
def flow_info(
    flow_db_id: flow_db_id_arg,
    job_id_flag: job_flow_id_flag_opt = False,
):
    """
    Provide detailed information on a Flow
    """
    db_id, jf_id = get_job_db_ids(flow_db_id, None)
    db_ids = job_ids = flow_ids = None
    if db_id is not None:
        db_ids = [db_id]
    elif job_id_flag:
        job_ids = [jf_id]
    else:
        flow_ids = [jf_id]

    with loading_spinner():
        jc = get_job_controller()

        flows_info = jc.get_flows_info(
            job_ids=job_ids,
            db_ids=db_ids,
            flow_ids=flow_ids,
            limit=1,
            full=True,
        )
    if not flows_info:
        exit_with_error_msg("No data matching the request")

    out_console.print(format_flow_info(flows_info[0]))


@app_flow.command()
def graph(
    flow_db_id: flow_db_id_arg,
    job_id_flag: job_flow_id_flag_opt = False,
    label: Annotated[
        Optional[str],
        typer.Option(
            "--label",
            "-l",
            help="The label used to identify the nodes",
        ),
    ] = "name",
    file_path: Annotated[
        Optional[str],
        typer.Option(
            "--path",
            "-p",
            help="If defined, the graph will be dumped to a file",
        ),
    ] = None,
    dash_plot: Annotated[
        bool,
        typer.Option(
            "--dash",
            "-d",
            help="Show the graph in a dash app",
        ),
    ] = False,
    print_mermaid: Annotated[
        bool,
        typer.Option(
            "--mermaid",
            "-m",
            help="Print the mermaid graph",
        ),
    ] = False,
):
    """
    Provide detailed information on a Flow
    """
    db_id, jf_id = get_job_db_ids(flow_db_id, None)
    db_ids = job_ids = flow_ids = None
    if db_id is not None:
        db_ids = [db_id]
    elif job_id_flag:
        job_ids = [jf_id]
    else:
        flow_ids = [jf_id]

    with loading_spinner():
        jc = get_job_controller()

        flows_info = jc.get_flows_info(
            job_ids=job_ids,
            db_ids=db_ids,
            flow_ids=flow_ids,
            limit=1,
            full=True,
        )
    if not flows_info:
        exit_with_error_msg("No data matching the request")

    if print_mermaid:
        from jobflow_remote.jobs.graph import get_mermaid

        print(get_mermaid(flows_info[0]))

    if dash_plot:
        plot_dash(flows_info[0])
    else:
        plt = draw_graph(get_graph(flows_info[0], label=label))
        if file_path:
            plt.savefig(file_path)
        else:
            plt.show()



"""A series of toy workflows that can be used for testing."""

from typing import Callable, Optional, Union

from jobflow import job


@job
def add(a, b):
    """Adds two numbers together and writes the answer to a file."""
    return a + b


@job
def always_fails():
    """A job that always fails."""
    raise RuntimeError("This job failed.")


@job
def write_file(n):
    with open("results.txt", "w") as f:
        f.write(str(n))
    return


@job
def arithmetic(
    a: Union[float, list[float]],
    b: Union[float, list[float]],
    op: Optional[Callable] = None,
) -> Optional[float]:
    if op:
        return op(a, b)

    return None


@job
def check_env_var() -> str:
    import os

    return os.environ.get("TESTING_ENV_VAR", "unset")



from __future__ import annotations

import functools
import json
import logging
import uuid
from contextlib import contextmanager
from datetime import datetime, timedelta
from enum import Enum
from typing import TYPE_CHECKING, Callable

import typer
from click import ClickException
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Confirm
from rich.text import Text

from jobflow_remote import ConfigManager, JobController
from jobflow_remote.config.base import ProjectUndefined
from jobflow_remote.jobs.daemon import DaemonError, DaemonManager, DaemonStatus
from jobflow_remote.jobs.state import JobState

if TYPE_CHECKING:
    from cProfile import Profile

logger = logging.getLogger(__name__)


err_console = Console(stderr=True)
out_console = Console()


fmt_datetime = "%Y-%m-%d %H:%M"


# shared instances of the config manager and job controller, to avoid parsing
# the files multiple times. Needs to be initialized with the
# initialize_config_manager function.
_shared_config_manager: ConfigManager | None = None
_shared_job_controller: JobController | None = None

_profiler: Profile | None = None


def initialize_config_manager(*args, **kwargs):
    global _shared_config_manager
    _shared_config_manager = ConfigManager(*args, **kwargs)


def get_config_manager() -> ConfigManager:
    global _shared_config_manager
    if not _shared_config_manager:
        raise RuntimeError("The shared config manager needs to be initialized")
    return _shared_config_manager


def get_job_controller():
    global _shared_job_controller
    if _shared_job_controller is None:
        cm = get_config_manager()
        jc = JobController.from_project(cm.get_project())
        _shared_job_controller = jc

    return _shared_job_controller


def cleanup_job_controller():
    global _shared_job_controller
    if _shared_job_controller is not None:
        _shared_job_controller.close()


def start_profiling():
    global _profiler
    from cProfile import Profile

    _profiler = Profile()
    _profiler.enable()


def complete_profiling():
    global _profiler

    _profiler.disable()
    import pstats

    stats = pstats.Stats(_profiler).sort_stats("cumtime")
    stats.print_stats()


class SortOption(Enum):
    CREATED_ON = "created_on"
    UPDATED_ON = "updated_on"
    DB_ID = "db_id"


class SerializeFileFormat(Enum):
    JSON = "json"
    YAML = "yaml"
    TOML = "toml"


class ReprStr(str):
    """
    Helper class that overrides the standard __repr__ to return the string itself
    and not its repr().
    Used mainly to allow printing of strings with newlines instead of '\n' when
    repr is used in rich.
    """

    def __repr__(self):
        return self


def exit_with_error_msg(message: str, code: int = 1, **kwargs):
    kwargs.setdefault("style", "red")
    err_console.print(message, **kwargs)
    raise typer.Exit(code)


def exit_with_warning_msg(message: str, code: int = 0, **kwargs):
    kwargs.setdefault("style", "gold1")
    err_console.print(message, **kwargs)
    raise typer.Exit(code)


def print_success_msg(message: str = "operation completed", **kwargs):
    kwargs.setdefault("style", "green")
    out_console.print(message, **kwargs)


def check_incompatible_opt(d: dict):
    not_none = []
    for k, v in d.items():
        if v:
            not_none.append(k)

    if len(not_none) > 1:
        options_list = ", ".join(not_none)
        exit_with_error_msg(f"Options {options_list} are incompatible")


def check_at_least_one_opt(d: dict):
    not_none = []
    for k, v in d.items():
        if v:
            not_none.append(k)

    if len(not_none) > 1:
        options_list = ", ".join(d.keys())
        exit_with_error_msg(
            f"At least one of the options {options_list} should be defined"
        )


def check_only_one_opt(d: dict):
    not_none = []
    for k, v in d.items():
        if v:
            not_none.append(k)

    if len(not_none) != 1:
        options_list = ", ".join(d.keys())
        exit_with_error_msg(
            f"One and only one of the options {options_list} should be defined"
        )


@contextmanager
def loading_spinner(processing: bool = True):
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        if processing:
            progress.add_task(description="Processing...", total=None)
        yield progress


def get_job_db_ids(job_db_id: str, job_index: int | None):
    if check_valid_uuid(job_db_id, raise_on_error=False):
        db_id = None
        job_id = job_db_id
    else:
        db_id = job_db_id
        job_id = None

    if job_index and db_id is not None:
        out_console.print(
            "The index is defined even if an integer is passed as an ID. Will be ignored",
            style="yellow",
        )

    return db_id, job_id


def get_job_ids_indexes(job_ids: list[str] | None) -> list[tuple[str, int]] | None:
    if not job_ids:
        return None
    job_ids_indexes = []
    for j in job_ids:
        split = j.split(":")
        if len(split) != 2 or not split[1].isnumeric():
            raise typer.BadParameter(
                "The job id should be in the format UUID:INDEX "
                "(e.g. e1d66c4f-81db-4fff-bda2-2bf1d79d5961:2). "
                f"Wrong format for {j}"
            )
        check_valid_uuid(split[0])
        job_ids_indexes.append((split[0], int(split[1])))

    return job_ids_indexes


def cli_error_handler(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (typer.Exit, typer.Abort, ClickException):
            raise  # Do not capture click or typer exceptions
        except ProjectUndefined:
            exit_with_error_msg(
                "The active project could not be determined and it is required to execute this command"
            )
        except Exception as e:
            from jobflow_remote import SETTINGS

            if SETTINGS.cli_full_exc:
                raise  # Reraise exceptions to print the full stacktrace
            else:
                exit_with_error_msg(
                    f"An Error occurred during the command execution: {e.__class__.__name__} {getattr(e, 'message', str(e))}"
                )

    return wrapper


def check_valid_uuid(uuid_str, raise_on_error: bool = True) -> bool:
    try:
        uuid_obj = uuid.UUID(uuid_str)
        if str(uuid_obj) == uuid_str:
            return True
    except ValueError:
        pass

    if raise_on_error:
        raise typer.BadParameter(f"UUID {uuid_str} is in the wrong format.")
    else:
        return False


def str_to_dict(string: str | None) -> dict | None:
    if not string:
        return None

    try:
        dictionary = json.loads(string)
    except json.JSONDecodeError:
        dictionary = {}
        for chunk in string.split(","):
            split = chunk.split("=")
            if len(split) != 2:
                raise typer.BadParameter(
                    f"Wrong format for dictionary-like field {string}"
                )

            dictionary[split[0]] = split[1]

    return dictionary


def get_start_date(start_date: datetime | None, days: int | None, hours: int | None):
    if start_date and (start_date.year, start_date.month, start_date.day) == (
        1900,
        1,
        1,
    ):
        now = datetime.now()
        start_date = start_date.replace(year=now.year, month=now.month, day=now.day)
        if start_date > now:
            start_date = start_date - timedelta(days=1)
    elif days:
        start_date = datetime.now() - timedelta(days=days)
    elif hours:
        start_date = datetime.now() - timedelta(hours=hours)

    return start_date


def execute_multi_jobs_cmd(
    single_cmd: Callable,
    multi_cmd: Callable,
    job_db_id: str | None = None,
    job_index: int | None = None,
    job_ids: list[str] | None = None,
    db_ids: str | list[str] | None = None,
    flow_ids: str | list[str] | None = None,
    state: JobState | None = None,
    start_date: datetime | None = None,
    end_date: datetime | None = None,
    name: str | None = None,
    metadata: str | None = None,
    days: int | None = None,
    hours: int | None = None,
    verbosity: int = 0,
    raise_on_error: bool = False,
    **kwargs,
):
    query_values = [
        job_ids,
        db_ids,
        flow_ids,
        state,
        start_date,
        end_date,
        name,
        metadata,
        days,
        hours,
    ]
    try:
        if job_db_id is not None:
            if any(query_values):
                msg = "If job_db_id is defined all the other query options should be disabled"
                exit_with_error_msg(msg)
            db_id, job_id = get_job_db_ids(job_db_id, job_index)
            with loading_spinner():
                modified_ids = single_cmd(
                    job_id=job_id, job_index=job_index, db_id=db_id, **kwargs
                )
                if not modified_ids:
                    exit_with_error_msg("Could not perform the requested operation")
        else:
            check_incompatible_opt(
                {"start_date": start_date, "days": days, "hours": hours}
            )
            check_incompatible_opt({"end_date": end_date, "days": days, "hours": hours})
            metadata_dict = str_to_dict(metadata)

            job_ids_indexes = get_job_ids_indexes(job_ids)
            start_date = get_start_date(start_date, days, hours)

            if not any(
                (
                    job_ids_indexes,
                    db_ids,
                    flow_ids,
                    state,
                    start_date,
                    end_date,
                    name,
                    metadata,
                )
            ):
                text = Text.from_markup(
                    "[yellow]No filter has been set. This will apply the change to all "
                    "the jobs in the DB. Proceed anyway?[/yellow]"
                )

                confirmed = Confirm.ask(text, default=False)
                if not confirmed:
                    raise typer.Exit(0)

            with loading_spinner():
                modified_ids = multi_cmd(
                    job_ids=job_ids_indexes,
                    db_ids=db_ids,
                    flow_ids=flow_ids,
                    state=state,
                    start_date=start_date,
                    end_date=end_date,
                    name=name,
                    metadata=metadata_dict,
                    raise_on_error=raise_on_error,
                    **kwargs,
                )

        if verbosity:
            print_success_msg(f"Operation completed. Modified jobs: {modified_ids}")
        else:
            print_success_msg(f"Operation completed: {len(modified_ids)} jobs modified")
    except Exception:
        logger.error("Error executing the operation", exc_info=True)


def check_stopped_runner(error: bool = True):
    cm = get_config_manager()
    dm = DaemonManager.from_project(cm.get_project())
    try:
        with loading_spinner(False) as progress:
            progress.add_task(description="Checking the Daemon status...", total=None)
            current_status = dm.check_status()

    except DaemonError as e:
        exit_with_error_msg(
            f"Error while checking the status of the daemon: {getattr(e, 'message', str(e))}"
        )
    if current_status not in (DaemonStatus.STOPPED, DaemonStatus.SHUT_DOWN):
        if error:
            exit_with_error_msg(
                f"The status of the daemon is {current_status.value}. "
                "The daemon should not be running while resetting the database"
            )
        else:
            text = Text.from_markup(
                "[red]The Runner is active. This operation may lead to "
                "inconsistencies in this case. Proceed anyway?[/red]"
            )

            confirmed = Confirm.ask(text, default=False)
            if not confirmed:
                raise typer.Exit(0)
